
#include <unittest++/UnitTest++.h>

#include "VKRecorderRoot.hh"
#include "VKRunAction.hh"
#include "VKPrimaryGeneratorAction.hh"

#include "G4RunManager.hh"
#include "G4UImanager.hh"
#include "G4SystemOfUnits.hh"
#include "G4ThreeVector.hh"

#include "TMath.h"


SUITE(VKPrimaryGeneratorTest) {

    class SetupFixture {
        public:
            SetupFixture()
            {
                // get pointer to run manager
                runManager = G4RunManager::GetRunManager();
                
                // get pointer to UI manager
                UImanager = G4UImanager::GetUIpointer();
                
                // get pointer to user run action
                const VKRunAction *runAction = 
                    static_cast<const VKRunAction*>(runManager->GetUserRunAction()); 
                
                // get pointer to recorder
                recorder = static_cast<const VKRecorderRoot*>(runAction->GetRecorder());
                
                // get pointer to primary generator
                primGenAction = static_cast<const VKPrimaryGeneratorAction*>    
                                        (runManager->GetUserPrimaryGeneratorAction());

                // remove all parts of the setup
                UImanager->ApplyCommand("/VK/detector/volumes/betaDetector false");
                UImanager->ApplyCommand("/VK/detector/volumes/magnet false");
                UImanager->ApplyCommand("/VK/detector/volumes/catcherFoil false");
                UImanager->ApplyCommand("/VK/detector/volumes/shield false");

                // do *not* store trajectory data
                UImanager->ApplyCommand("/VK/storeTrajectoryData false");

                // disable magnetic field
                UImanager->ApplyCommand("/VK/field/enable false");
                
                // set particle gun
                UImanager->ApplyCommand("/gps/direction 0 0 1");
                UImanager->ApplyCommand("/gps/position 0 0 0 cm");
                UImanager->ApplyCommand("/gps/particle e-");
                UImanager->ApplyCommand("/gps/energy 30 MeV");
            }
            G4UImanager *UImanager;
            G4RunManager *runManager;
            const VKRecorderRoot *recorder;
            const VKPrimaryGeneratorAction *primGenAction;
    };

    TEST_FIXTURE(SetupFixture, Sanity) {
        CHECK_EQUAL(1, 1);
    }
    
    TEST_FIXTURE(SetupFixture, BetaSpectrumTest) {
        // enable beta spectrum and set end point
        UImanager->ApplyCommand("/VK/source/betaSpectrum true");
        UImanager->ApplyCommand("/VK/source/betaEndPoint 5.5 MeV");
        // run 20 events and check that energy is within bounds
        for (size_t i=0; i<20; i++) {
            UImanager->ApplyCommand("/run/beamOn 1");
            G4double edep = recorder -> GetEkinso();
            G4bool aboveZero = edep/MeV > 0;
            G4bool belowEndPoint = edep/MeV <= 5.5;
            CHECK_EQUAL(true, aboveZero); 
            CHECK_EQUAL(true, belowEndPoint); 
        }
        // disable beta spectrum 
        UImanager->ApplyCommand("/VK/source/betaSpectrum false");
    }

    TEST_FIXTURE(SetupFixture, CRMuonAngDistTest) {
        // enable CR muon ang dist
        UImanager->ApplyCommand("/VK/source/cosmicRayAngDist true");
        // set sampling bounds
        UImanager->ApplyCommand("/VK/source/cosmicRayMinZenithAngle 50 deg");
        UImanager->ApplyCommand("/VK/source/cosmicRayMaxZenithAngle 60 deg");
		// set power
		UImanager->ApplyCommand("/VK/source/cosmicRayPower 2.85");
        // unit vector
        G4ThreeVector n;
        // run 20 events and check that angle is within bounds
        for (size_t i=0; i<20; i++) {
            UImanager->ApplyCommand("/run/beamOn 1");
            G4double theta = recorder -> GetThetaso();
            G4double phi   = recorder -> GetPhiso();
            n.setRThetaPhi(1, theta, phi);
            double d = n.dot(G4ThreeVector( 0, -1, 0));
            double z = acos(d) *180./TMath::Pi();
            G4bool aboveLower = z >= 50;
            G4bool belowUpper = z <= 60;
            CHECK_EQUAL(true, aboveLower); 
            CHECK_EQUAL(true, belowUpper); 
        }
        // check coverage
        double cov = primGenAction -> GetCoverage();
        CHECK_CLOSE(0.113, cov, 1e-3);
        // disable CR muon ang dist
        UImanager->ApplyCommand("/VK/source/cosmicRayAngDist false");
    }

    TEST_FIXTURE(SetupFixture, SamplingBoundsForBetaSpectrumTest) {
        // enable beta spectrum and set end point
        UImanager->ApplyCommand("/VK/source/betaSpectrum true");
        UImanager->ApplyCommand("/VK/source/betaEndPoint 5.5 MeV");
        // set bounds for sampling
        UImanager->ApplyCommand("/VK/source/betaMinEnergy 2.10 MeV");
        UImanager->ApplyCommand("/VK/source/betaMaxEnergy 2.30 MeV");
        // run 20 events and check that energy is within bounds
        for (size_t i=0; i<20; i++) {
            UImanager->ApplyCommand("/run/beamOn 1");
            G4double edep = recorder -> GetEkinso();
            G4bool aboveLower= edep/MeV >= 2.10;
            G4bool belowUpper = edep/MeV <= 2.30;
            CHECK_EQUAL(true, aboveLower); 
            CHECK_EQUAL(true, belowUpper); 
        }
        // disable beta spectrum 
        UImanager->ApplyCommand("/VK/source/betaSpectrum false");
    }
    
    TEST_FIXTURE(SetupFixture, DummyTest) {
        // enable beta spectrum and set end point
        UImanager->ApplyCommand("/VK/source/betaSpectrum true");
        UImanager->ApplyCommand("/VK/source/betaEndPoint 5.5 MeV");
        // set bounds for sampling
        UImanager->ApplyCommand("/VK/source/betaMinEnergy 2.10 MeV");
        UImanager->ApplyCommand("/VK/source/betaMaxEnergy 2.30 MeV");
        // run 20 events and check that energy is within bounds
        for (size_t i=0; i<20; i++) {
            UImanager->ApplyCommand("/run/beamOn 1");
            G4double edep = recorder -> GetEkinso();
            G4bool aboveLower= edep/MeV >= 2.10;
            G4bool belowUpper = edep/MeV <= 2.30;
            CHECK_EQUAL(true, aboveLower); 
            CHECK_EQUAL(true, belowUpper); 
        }
        // disable beta spectrum 
        UImanager->ApplyCommand("/VK/source/betaSpectrum false");
    }    
   
    TEST_FIXTURE(SetupFixture, CoverageTest) {
        // enable beta spectrum and set end point
        UImanager->ApplyCommand("/VK/source/betaSpectrum true");
        UImanager->ApplyCommand("/VK/source/betaEndPoint 7.00 MeV");
        // set bounds for sampling
        UImanager->ApplyCommand("/VK/source/betaMinEnergy 5.40 MeV");
        UImanager->ApplyCommand("/VK/source/betaMaxEnergy 7.00 MeV");
        // get coverage
        double cov = primGenAction -> GetCoverage();
        CHECK_CLOSE(0.070, cov, 1e-3);
        // reduce angular coverage
//        UImanager->ApplyCommand("/gps/ang/type iso");
        UImanager->ApplyCommand("/gps/ang/mintheta  90 deg");
        UImanager->ApplyCommand("/gps/ang/maxtheta 180 deg");
        cov = primGenAction -> GetCoverage();
        CHECK_CLOSE(0.035, cov, 1e-3);
        // disable beta spectrum 
        UImanager->ApplyCommand("/VK/source/betaSpectrum false");
    }    

    TEST_FIXTURE(SetupFixture, SetSourceParametersTest) {
        // specify source position
        UImanager->ApplyCommand("/gps/position -10 30 25 mm");
        // enable disk-shaped source
        UImanager->ApplyCommand("/VK/source/diskShapedSource true");
        // specify source dimensions and orientation
        UImanager->ApplyCommand("/VK/source/innerDiameter 2 mm");
        UImanager->ApplyCommand("/VK/source/outerDiameter 10 mm");
        UImanager->ApplyCommand("/VK/source/surfaceNormal 0 0 1");
        // run 20 events and check start position
        for (size_t i=0; i<20; i++) {
            UImanager->ApplyCommand("/run/beamOn 1");
            double x = recorder -> GetXso() /mm;
            double y = recorder -> GetYso() /mm;
            double z = recorder -> GetZso() /mm;
            double r = pow(pow(x+10, 2)+pow(y-30, 2)+pow(z-25, 2), 0.5);  // mm
            G4bool minBound = r >= 1;
            G4bool maxBound = r <= 5;
            CHECK_EQUAL(true, minBound); 
            CHECK_EQUAL(true, maxBound); 
            CHECK_CLOSE(25, z, 1e-3);
        }
        // try horizontal source
        UImanager->ApplyCommand("/VK/source/innerDiameter 0 mm");
        UImanager->ApplyCommand("/VK/source/outerDiameter 200 mm");
        UImanager->ApplyCommand("/VK/source/surfaceNormal 0 1 0");
        // run 20 events and check start position
        for (size_t i=0; i<20; i++) {
            UImanager->ApplyCommand("/run/beamOn 1");
            double x = recorder -> GetXso() /mm;
            double y = recorder -> GetYso() /mm;
            double z = recorder -> GetZso() /mm;
            double r = pow(pow(x+10, 2)+pow(y-30, 2)+pow(z-25, 2), 0.5);  // mm
            G4bool minBound = r >= 0;
            G4bool maxBound = r <= 200;
            CHECK_EQUAL(true, minBound); 
            CHECK_EQUAL(true, maxBound); 
            CHECK_CLOSE(30, y, 1e-3);
        }        
        // disable disk-shaped source
        UImanager->ApplyCommand("/VK/source/diskShapedSource false");
    }
}
